from mpl_toolkits.mplot3d import proj3d
import numpy, importlib,sys


def orthogonal_proj(zfront, zback):
    a = (zfront+zback)/(zfront-zback)
    b = -2*(zfront*zback)/(zfront-zback)
    return numpy.array([[1,0,0,0],
                        [0,1,0,0],
                        [0,0,a,b],
                        [0,0,0,zback]])
# proj3d.persp_transformation = orthogonal_proj
def set_perspective(b=False):
    importlib.reload(proj3d)
    if not b :
        print('setting persp_transformation as orthogonal')
        proj3d.persp_transformation = orthogonal_proj

class handler_3d:
    def __init__(self,fig,e=[1,0,0],persp=False,xm=None):
        self.dazim=5
        self.delev=5
        self.fig = fig
        self.fig.canvas.mpl_connect('key_press_event', self)
        self.e=e
        xm0 = round(fig.canvas.figure.axes[0].get_xlim()[1])
        self.xm0 = xm0
        self.xm  = xm0
        set_perspective(b=persp)
    def __call__(self,event):
        # print(event.key)
        ax = event.canvas.figure.axes[0];
        # dx,dy,dz = 0,0,0
        if   event.key == '1'     : ax.elev,ax.azim = 0,0
        elif event.key == '2'     : ax.elev,ax.azim = 0,90
        elif event.key == '3'     : ax.elev,ax.azim = 90,90
        elif event.key == '4'     : ax.elev,ax.azim = 180,0
        elif event.key == '5'     : ax.elev,ax.azim = 0,-90
        elif event.key == '6'     : ax.elev,ax.azim = -90,-90
        elif event.key == 'up'    : ax.elev-=self.delev
        elif event.key == 'down'  : ax.elev+=self.delev
        elif event.key == 'left'  : ax.azim-=self.dazim
        elif event.key == 'right' : ax.azim+=self.dazim
        elif event.key == 'ctrl+up'   : self.delev+=5
        elif event.key == 'ctrl+down' : self.delev = max(5,self.delev-5)
        elif event.key == 'ctrl+left' : self.dazim = max(5,self.dazim-5)
        elif event.key == 'ctrl+right': self.dazim+=5
        elif event.key == 'ctrl+r'    : self.delev,self.dazim=5,5
        elif event.key == 'pageup'   : self.xm /=2
        elif event.key == 'pagedown' : self.xm *=2
        elif event.key == 'r'        : self.xm = self.xm0
        if 'ctrl' in event.key :
            print('delev=%d,dazim=%d' %(self.delev,self.dazim))
        if event.key in '123456updownleftright':
            print('elev=%d,azim=%d' %(ax.elev,ax.azim))
        if event.key in 'rpageuppagedown' :
            xm = self.xm
            ax.set_xlim([-xm,xm])
            ax.set_ylim([-xm,xm])
            ax.set_zlim([-xm,xm])
            print('xm : ',xm)
        event.canvas.draw()



'''Generic mathematical statistical functions for semiconductors
'''
import time
import scipy.special as spe
import numpy as np
import mpmath as mpm
from . import displayStandards as dsp
# from integrationRect import *
from . import newtonRaphson as nwt


##### def : Maxwell-Boltzmann Statistics
def MaxwellBoltz(x):
    return np.exp(-x)

##### def Fermi-Dirac statistics Th
def FermiDirac(x):
    return 1/(1+np.exp(x))
def FD_int(etas,j,dps=5):
    mpm.mp.dps = dps; mpm.mp.pretty = True;
    etas = np.array([etas]) if np.isscalar(etas) else np.array(etas)
    Li      = [mpm.polylog(j+1,-np.exp(float(eta))).real for eta in etas]
    FDint   = -spe.gamma(j+1)*np.array(Li)
    return FDint
def FermiDiracInt_Th(etas,dps=10):
    return FD_int(etas,0.5,dps)
def dFermiDiracInt_Th(etas,dps=10):
    return spe.gamma(1.5)/spe.gamma(0.5)*FD_int(etas,-0.5,dps)

def FermiDiracIntInv0(x,eta0=0,dps=5):
    f  = lambda eta: FermiDiracInt_Th(eta,dps) - x
    df = lambda eta:dFermiDiracInt_Th(eta,dps)
    eta0 = nwt.newtonRaphson1D(f,df,eta0,tol=0.0001,nMax=10)
    #eta0 = bis.bissection2(f,-2,2,tol=0.001,Nmax=100)
    #eta0 = nwt.newtonRaphson1Ddebug(f,df,eta0,np.linspace(0,3,10), tol=0.0001,nMax=10)
    eta0 = eta0[0]
    return eta0

def FermiDiracIntInv(x,eta0=0,dps=5):
    x = np.array([x]) if np.isscalar(x) else np.array(x)
    N = x.size
    etas  = np.zeros((N))
    for i in range(N):
        eta0 = FermiDiracIntInv0(x[i],eta0,dps); #printeta0
        etas[i] = eta0
    return etas



###### Fermi Dirac numerical integrations
def FermiDiracInt(eta,dx=0.001,xMax=100):
    eta = np.array([eta]) if np.isscalar(eta) else np.array(eta)
    N       = eta.size;
    FDint  = np.zeros((N))
    for i in range(N):
        X        = np.arange(0,eta[i]+xMax,dx)
        FDint[i] = sum(np.sqrt(X)*FermiDirac(X-eta[i]) )*dx
    return FDint
def dFermiDiracInt(eta,dx=0.001,xMax=100):
    eta = np.array([eta]) if np.isscalar(eta) else np.array(eta)
    N       = eta.size;
    dFDint  = np.zeros((N))
    for i in range(N):
        X = np.arange(0,eta[i]+xMax,dx)
        dFDint[i] = sum(np.sqrt(X)*np.exp(X-eta[i])*pow(FermiDirac(X-eta[i]),2))*dx
    return dFDint

def dFermiDiracIntInv(x,eta0=0,dps=5):
    return 1/dFermiDiracInt_Th(FermiDiracIntInv(x,eta0,dps))


#### def : Misc:
dummyLambda = lambda E:np.ones((E.size))
#####################################################################
##### def : display
def displayLi():
    mpm.mp.dps = 15; mpm.mp.pretty = True;
    etas = np.linspace(-2,0.8,100)
    Lim2 = np.array([mpm.polylog(-2,eta) for eta in etas])
    Lim1 = np.array([mpm.polylog(-1,eta) for eta in etas])
    Li0  = np.array([mpm.polylog( 0,eta) for eta in etas])
    Li1  = np.array([mpm.polylog( 1,eta) for eta in etas])
    Li2  = np.array([mpm.polylog( 2,eta) for eta in etas])
    xlabs = ['$\eta$','$Li$']
    plots = [[etas,Lim2,'r','$Li_{-2}(x)$'],
             [etas,Lim1,'y','$Li_{-1}(x)$'],
             [etas,Li0 ,'g','$Li_0(x)$'],
             [etas,Li1 ,'b','$Li_1(x)$'],
             [etas,Li2 ,'m','$Li_2(x)$']]
    stdDispPlt(plots,xlabs,xlims=[-2,1,-1,1])
def display_FermiInt():
    eta = np.linspace(-2,2,100)

    t0 = time.clock()
    FDint = FermiDiracInt(eta,dx=0.001,xMax=50)
    print('My computation %f ' %(time.clock()-t0));t0 = time.clock();
    FDint_th1 = FermiDiracInt_Th(eta,dps=4)
    print('Th computation %f ' %(time.clock()-t0));t0 = time.clock();

    plots = [[eta,FDint,'b','$num$'],
             [eta,FDint_th1 ,'r','$th_1$' ],]
    xlabs = ['$\eta$','$F_{1/2}(\eta)}$']
    stdDispPlt(plots,xlabs)

def display_dFermiInt(opt=0,timeOpt=0):
    eta = np.linspace(-2,2,100)
    FDint_Th    = lambda x:FermiDiracInt_Th(x)
    dFDint_Th   = lambda x:dFermiDiracInt_Th(x)
    FDint_Num   = lambda x:FermiDiracInt(x)
    dFDint_Num  = lambda x:dFermiDiracInt(x)
    if opt:
        if timeOpt:
            t0 = time.clock()
            FDint_num = FermiDiracInt(eta,dx=0.01,xMax=50)
            print('My computation %f ' %(time.clock()-t0));t0 = time.clock();
            FDint_th = FermiDiracInt_Th(eta,dps=10)
            print('Th computation %f ' %(time.clock()-t0));t0 = time.clock();
        else :
            FDint_num = dFDint_Num(eta)
            FDint_th  = dFDint_Th(eta)
        plots = [[eta, FDint_num  ,'b','$num$'],
                 [eta, FDint_th,'r','$th$' ],]
        xlabs = ['$\eta$','$\partial_{\eta}F_{1/2}(\eta)}$']
        stdDispPlt(plots,xlabs)
    else :
        displayDerivative(eta,FDint_Th  ,dFDint_Th ,['\eta','F_{-1/2}^{ana} '],showOpt=0)
        #displayDerivative(eta,FDint_Num,dFDint_Num,['\eta','F_{-1/2}^{int}'],showOpt=0)
        plt.show()


#####################################################################
##### def : display
# inversions
def testFermiInversion():
    x,eta0 = 0,0
    eta0 = FermiDiracIntInv0(x,eta0);

    f  = lambda eta:FermiDiracInt_Th(eta) - x
    df = lambda eta:dFermiDiracInt_Th(eta)
    etas = np.linspace(-2,2,50)
    plots = [[etas,f(etas) , 'b-+' ,'$f_x(\eta)$'],
             [eta0,f(eta0) ,'bs'   ,'$\eta_0$'],]
             #[etas,df(etas), 'r--' ,'$\partial_{\eta} f_x(\eta)$'],
    stdDispPlt(plots,['$\eta $','$f_x(\eta)$'],axPos=[0.15,0.12,0.8,0.8])

def testFermiInversions():
    #x = np.concatenate((np.linspace(0,0.5,50),np.linspace(0.5,1,10)) )
    x = np.linspace(0.1,1,100)
    eta0   = 0
    etas   = FermiDiracIntInv(x,eta0);
    x0     = FermiDiracInt_Th(etas)
    plots  = [[x,etas   , 'r-+' ,'$F^{-1}(x)$'],
              [etas,x   , 'm-+' ,'$F(\eta)$'],
              [x,x0     , 'k-+' ,'$F(F^{-1}(x))=x$'],]
    stdDispPlt(plots,['$x $','$F_{1/2}^{-1}(x)$'])

def testdFermiIntInv():
    x = np.concatenate((np.linspace(0.1,0.5,50),np.linspace(0.5,1,10)) )
    eta0 = 0
    FDintInv  = lambda x:FermiDiracIntInv(x,eta0)
    dFDintInv = lambda x:dFermiDiracIntInv(x,eta0)
    displayDerivative(x,FDintInv,dFDintInv,['\eta','F_{-1/2}(x)'])


if __name__=='__main__':
    #displayLi()
    #display_FermiInt()
    #display_dFermiInt()
    #testFermiInversion()
    #testFermiInversions()
    testdFermiIntInv()
    print("Statistics compiled")



import numpy as np
import pandas as pd
import pickle,os


from GaAs import mat as GaAs
from InAs import mat as InAs
from AlAs import mat as AlAs
from Si import mat as Si
from Ge import mat as Ge

##add materials here
mats = {'GaAs':GaAs,'AlAs':AlAs,'InAs':InAs, 'Si':Si,'Ge':Ge}

sc_cond   = ['Eg','Xi','eps','Nc','Nv']
masses    = ['me','mh','mhh','mlh','mso']
lattice   = ['a0']
luttinger = ['g1','g2','g3']
stiffness = ['C11','C12','C44']
optical   = ['n_ref']
other     = ['a','b','d']

cols = sc_cond+masses+lattice+luttinger+stiffness+optical+other
df = pd.DataFrame(columns=cols)

for mat_name,mat in mats.items():
    df.loc[mat_name] = [np.nan]*len(cols)
    for k,v in mat.items():df.loc[mat_name][k]=v

filename = os.path.dirname(__file__)+'/materials.pkl'
df.to_pickle(filename)
print(filename+ ' saved')



from FourierUtils import*
from displayStandards import*
from scipy import signal


#__all__=['base_test']
def base_test():
    ''' Show the Fourier transform of a square in  '''
    Tf,dt = 10, 0.01
    T,nT,duty = 1,1,1

    yfunc = lambda t:np.array(squareBis(t,T,nT,duty),dtype=complex)
    t,y,f,Y = Fourier(yfunc,Tf=Tf,dt=dt,
                      fopt='',dOpt='fts',pOpt='RIYP',
                      labOpts=['x','k'],
                      xlims={'t':[-10,10,-0.1,1.1],'f':[]})

def test_showCase():
    Tf,dt = 10000, 0.1
    T,nT,duty = 10,20,0.2

    t = np.arange(0,Tf+dt,dt) - Tf/2
    fig,ax = plt.subplots(2,1)

    y = squareBis(t,10,50,duty=0.2)
    y0 = squareBis(t,1000,1,duty=0.5)
    yF = squareBis(t,10,100,duty=0.2)
    Ts  = np.arange(0,Tf+dt,T) - Tf/2; nTs = Ts.size;Ts = np.array([Ts,Ts])
    #ax[0].plot(Ts,[0,1],'g--',label=''      ,alpha=0.25,linewidth=0.5)
    ax[0].plot(t,yF,'b'  ,label='periodic'  ,alpha=0.15,linewidth=0.75)
    ax[0].plot(t,y,'b'   ,label='$f(x)$'    ,alpha=1.0,linewidth=0.75)
    ax[0].plot(t,y0,'r',label='$y_0$'       ,alpha=0.5,linewidth=2.0)

    standardDisplay(ax[0],['$x$','$f$'],opt='q',axPos=22,xylims=[-500,500,0,1.2]) #,xyTickLabs=[[''],['']])

    f,Y = get_FFT(t,y,dt)
    f0,Y0 = get_FFT(t,y0,dt)

    Y = abs(Y)**2;          Ym=Y.max()
    Y0 = abs(Y0)**2;        Y0*=Ym/Y0.max()

    x0  = np.arange(-5,5.1,0.1)
    FF = np.sinc(2*f)**2*Ym
    FF0 = np.sinc(2*x0)**2*Ym
    ax[1].plot(f,FF    ,'b--' ,label='$sinc^2$',alpha=0.4,linewidth=3.0)
    ax[1].plot(x0,FF0  ,'bo'  ,label=''        ,alpha=0.4,markersize=15)
    ax[1].plot(f,abs(Y),'b'   ,label='$F(f)$'  ,alpha=1.0,linewidth=2.0)
    ax[1].plot(f0,Y0   ,'r'   ,label='$Y_0$'   ,alpha=1.0,linewidth=1.0)

    standardDisplay(ax[1],['$X$','$|F|^2$'],opt='q',axPos=21,xylims=[-1,1,0,1.1*10000])
    ax[0].xaxis.set_ticklabels([])
    ax[0].yaxis.set_ticklabels([])
    ax[1].xaxis.set_ticklabels([])
    ax[1].yaxis.set_ticklabels([])
    plt.show()

if __name__=="__main__":
    base_test()
    #test_showCase()



import importlib as imp
import numpy as np
import matplotlib.pyplot as plt
from numpy import pi
import scipy.optimize as sci_opt
import utils.newtonRaphson as nwt
import utils.displayStandards as dsp
imp.reload(nwt)
imp.reload(dsp)

def test_newton_rec():
    f = lambda y,x : np.tan(pi/2*(y-x))
    df = lambda y,x : pi/2*(1+f(y,x)**2)
    cond = lambda x0,x1,y0,y1 : y1>y0

    x,y0 = np.array([0,1.6]),0.01
    y0s = nwt.newton_recursive(f,df,x,y0,cond,Ncut=3,npts=3,tol=1e-5,nMax=100)
    f0s = np.array([f(y0,x0) for y0,x0 in zip(y0s,x)])

    #plot
    y,cs = np.linspace(0,2,1000),dsp.getCs('jet',x.size)
    plts = [[y,f(y,x0),c] for x0,c in zip(x,cs)]
    plts += [[y0s[i],f0s[i],[cs[i],'s']] for i in range(x.size)]
    plts += [[[0,y.max()],[0,0],'k']]
    dsp.stddisp(plts,labs=['$y$','$f(y;x)$'],lw=2,xylims=['y',-20,20],ms=5)


def test_newton_cplx_rec():
    f  = lambda z,z0 : np.sqrt(z**2 - z0**2)
    df = lambda z,z0 : z/f(z,z0)

    z0s = np.array([1+0.2J,1.5+0.15J,2.2+0.1J,3+0J])
    z0_nwt = nwt.newton_recursive(f,df,z0s,y0=1+0J,tol=1e-5,nMax=10,v=1)

    plts = [[z0s.imag,z0_nwt.imag,'rs-','im'],
            [z0s.real,z0_nwt.real,'bs-','re']]
    dsp.stddisp(plts)

def test_newton_cplx(opts='nm'):
    f  = lambda z,z0 : np.sqrt(z**2 - z0**2)
    df = lambda z,z0 : z/f(z,z0)
    z0 = 1+0.2J

    #newton
    z0_nwt = 0.7+0.1J
    if 'n' in opts:
        methods={'N':'Newton-CG','M':'Nelder-Mead','B':'BFGS','C':'CG'}
        method = methods['N']
        z0_nwt = nwt.newton_cplx(f,df,z0=z0_nwt,args=(z0,),tol=1e-5,nMax=10)[0]
        print(z0_nwt)

    #plot
    if 'm' in opts:
        npts = 100
        x,y = np.linspace(0,2,npts),np.linspace(-1,1,npts)
        X,Y = np.meshgrid(x,y)
        Z = X+1J*Y
        fz = np.abs(f(Z,z0))**2
        dF = lambda x,y:nwt.grad_fz2(f(x+1J*y,z0),df(x+1J*y,z0))

        x0,y0=z0_nwt.real,z0.imag
        plts = [[x0],[y0],'gs','$Z_0$']
        im = None #[X,Y,np.abs(fz)]
        fig,ax=dsp.stddisp(plots=plts,im=im,
            contour=[X,Y,fz,12],quiv=dF,
            imOpt='ch',pOpt='tG',lw=3,ms=10)



######################################################################

plt.close('all')
# test_newton_cplx()
test_newton_cplx_rec()
# x = nwt.newton(F,dF,x0,args=(z0,),tol=1e-5)
# test_newton_rec()
# test_newton_cplx()



#print('importing...')

# py standard library
#from math import*
import numpy as np
import matplotlib.pyplot as plt
import scipy.special as spe
import pandas as pd


# specific libraries
from . import displayStandards as dsp
from . import regionManager as regMg
from . import materialManager as matMg
from . import newtonRaphson as nwt
from . import bissection as bis
from . import commonInputOutput as iostd
from . import physicsConstants as cst
from . import glob_colors  as colors
from .glob_colors import *

def funcs(obj):
    import inspect
    functions = '\n'.join([f[0] for f in inspect.getmembers(obj,predicate=inspect.isfunction)])
    print(functions)



'''
Region Manager to assign regions to a 1D mesh and retrieve values
'''
import numpy as np
from . import materialManager as matMg

######
def set_region1D(x,xlims,**kwargs):
    '''define a 1D region from a 1D mesh x with boundaries xlims
    - x : 1D mesh (np.1darray)
    - xlims : region boundaries (do not need to lie within mesh)
    - kwargs : optional associated properties such as name, doping
    returns :\n
    - region : dict containing region info kwargs
        - x[region['idx']] : coordinates of region
        - region['lims']   : boundaries of region
    '''
    id0 = abs(x-xlims[0]).argmin()
    id1 = abs(x-xlims[1]).argmin()

    idx = slice(id0,id1+1)#(x[id1] in x))
    xlims = x[[id0,id1]]

    region = {'idx':idx,'lims':xlims}
    region.update(kwargs)
    return region

def get_region_params(reg,keys,T=300):
    '''Get region parameters
    Note : mat_name must exist for this
    Fetch from materialManager if keys not found in reg.keys()
    '''
    reg_keys = [k for k in keys if k in reg.keys()]
    mat_keys = list(np.setdiff1d(keys,reg_keys+['mat_name']))
    mat_vals = matMg.get_params(reg['mat_name'],T,mat_keys)
    reg_vals = [reg[k] for k in reg_keys]
    vals = dict(zip(mat_keys+reg_keys, list(mat_vals)+reg_vals))
    values = [vals[k] for k in keys]
    return values

def set_regions_arrays(x,regions,keys,T=300):
    '''Set properties arrays from regions info
    Inputs :\n
    - x : mesh
    - regions : list of region dictionaries
    - keys : The properties(Eg,Xi,...) to fill the arrays with
    - T : temperature
    Note : At the moment the values are assumed constants per region
    Returns : \n
    - reg_arrays : list of arrays filled with region values
    '''
    nkeys = len(keys)
    reg_arrays = [np.zeros(x.shape) for k in range(nkeys)];
    for reg in regions:
        #print('\t'+reg['mat_name'])
        values = get_region_params(reg,keys,T)
        idx = reg['idx']
        for reg_a,val in zip(reg_arrays,values):
            reg_a[idx] = val
    return reg_arrays



import numpy as np
import pandas as pd
import os
from .physicsConstants import kB,eV
from . import glob_colors as colors

materials = pd.read_pickle(os.path.dirname(__file__)+'/materials/materials.pkl')
_key_cols = materials.columns.values.tolist()
_key_funcs = ['Ncv','Ni','Neff']
_keys = _key_cols+_key_funcs


## functions to compute new properties from existing entries
get_NcNv = lambda mat,T: np.sqrt(mat['Nc']*mat['Nv'])
get_Ni   = lambda mat,T: get_NcNv(mat,T)*np.exp(-mat['Eg']/(2*(kB*T/eV)))
get_Neff = lambda mat,T: 2.5e19*mat['me']**1.5*(T/300)**1.5
funcs = dict(zip(_key_funcs,[get_NcNv,get_Ni,get_Neff]))

def get_params(mat_name,T,keys):
    ''' get properties of a material
    matName : str - Material name
    T       : Temperature (K)
    keys    : properties as available in materialManager._keys
    '''
    kT = kB*T/eV
    mat = materials.loc[mat_name]
    key_funcs = [k for k in keys if k in _key_funcs ]#;print(key_funcs)
    for kf in key_funcs : mat[kf] = funcs[kf](mat,T)
    values = mat[keys].values
    return values

def update_materials():
    import subprocess
    mat_path = os.path.dirname(__file__)+'/materials/'
    out=subprocess.check_output("python3 %smake_df_mat.py" %mat_path,shell=True).decode()
    print(colors.green+out+colors.black)
    materials = pd.read_pickle(mat_path+'materials.pkl')
# class Material:
#     def __init__(self,name):
#         vals = materials.loc[name].values
#         self.__dict__=dict(zip(cols,vals))
# get_NcNv = lambda mat: np.sqrt(mat.Nc*mat.Nv)
# get_Ni   = lambda mat: np.sqrt(mat.Nc*mat.Nv)*np.exp(-mat.Eg/(2*mat.kT))
# get_Neff = lambda mat: 2.5*e19*pow(mat.me,1.5)*pow(mat.T/300,1.5)

def computeIntrinsicDensity(Nc,Nv,Eg,T):
    Ni2 = Nc*Nv*np.exp(-Eg/(kB*T/eV))
    return np.sqrt(Ni2)

# def doping(x,Nd,Na):
#     N = len(x)
#     Ndx = np.zeros((N))
#     Nax = np.zeros((N))
#     for i in range(0,N):
#         if x[i]>0:
#             Ndx[i] = Nd
#         if x[i]<0:
#             Nax[i] = Na
#     return Ndx,Nax
#
# def computeEffectiveDensity(mc,T=300):
#     Neff = 2.5*e19*pow(me,1.5)*pow(T/300,1.5);
#     return Neff

linear_interp = lambda x,v1,v2 : x*v1 + (1-x)*v2
def getTernaryAlloy(mat1,mat2,x,keys=None,T=300):
    if not keys : keys = ['Eg','a0', 'g1','g2','g3', 'C11','C12','C44', 'a','b','d']
    mat1 = get_params(mat1,T,keys)
    mat2 = get_params(mat2,T,keys);#print(mat2)
    values = [linear_interp(x,v1,v2) for v1,v2 in zip(mat1,mat2)]
    return values


########################################################################
# tests
########################################################################
def _testTernaryAlloys(mat1='GaAs',mat2='AlAs',x=0.5):
    keys = ['Eg','a0', 'g1','g2','g3', 'C11','C12','C44', 'a','b','d']
    vals = getTernaryAlloy(mat1,mat2,x,keys)
    print('\n\t'+'AlGaAs\n',dict(zip(keys,vals)))

def _test_get_params(mat_name='GaAs',T=300):
    print('\n\t'+mat_name)
    keys=['Nc','Ni','Neff']; print(dict(zip(keys,get_params(mat_name,T,keys))))
    keys=['Eg','eps','Xi',]; print(dict(zip(keys,get_params(mat_name,T,keys))))

if __name__=='__main__':
    name = 'GaAs'
    _test_get_params(name)
    _testTernaryAlloys()



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from .glob_colors import*
from .displayStandards import saveFig
from matplotlib.widgets import RectangleSelector


###########################################################################
# Class : event Handler generic
class eventHandler:
    def __init__(self,fig,ax,path='./',name=''):
        self.cid  = fig.canvas.mpl_connect('key_press_event', self)
        self.fig  = fig
        self.ax   = ax
        self.RS = RectangleSelector(
            ax, self.line_select_callback, drawtype='box', useblit=True,
            button=[1, 3],spancoords='pixels', minspanx=20,minspany=20 )
        self.path,self.name = path,name
        self.selectBox = [0.0,0.0,0.0,0.0]
        self.params = {'paramTest':[0,1]}
        plt.rcParams['keymap.save'] = ''
    
    def __call__(self, event):
        #print ('key entered: ' + key)
        if event.key=='s':
            self.save()
        if event.key=='l':
            self.load()
        if event.key == '0':
            self.action0()
        if event.key in ['1','2','3','4','5','6','7','8','9']:
            self.displayAx(int(event.key))
        if event.key == 'enter':
            self.returnAction()
        if event.key == 'backspace':
            self.backspaceAction()
        if event.key == ' ':
            self.spaceAction()
        if event.key in ['up','right','left','down']:
            self.arrowActions(event.key)
        if event.key in ['a','q']:
            self.activateBox(event.key)
        if event.key in 'zertyuiopdfghjkmwxcvbn':
            self.keyCallBack(event.key)
        self.update()
        
    #can be overriden
    def save(self):
        saveFig(self.path+self.name, self.ax,fmt='png')
    def load(self):
        print('Base load ')
    def update(self):
        print('Base update')
    
    def keyCallBack(self,key):
        print(key)
    def activateBox(self,key):
        if key == 'q':
            self.activateRS(0)
        if key == 'a':
            self.activateRS(1)
    def activateRS(self,b):
        if b==0:
            print(' RectangleSelector deactivated.')
            self.RS.set_active(False)
        if b==1:
            print(' RectangleSelector activated.')
            self.RS.set_active(True)
            
    def line_select_callback(self,eclick, erelease):
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        self.selectBox = [x1,x2,y1,y2] 
        print(self.selectBox)

    def displayAx(self,i):
        print('base display ' + str(i))
    def action0(self):
        print('base action0')
    def returnAction(self):
        print('base returnAction')
    def backspaceAction(self):
        print('base backspaceAction')
    def spaceAction(self):
        print('base spaceAction')
    def arrowActions(self,key):
        print('base arrowAction ' + key)
        
def arrowActionsParam(v,dv,key):
    if key == 'up':
        v = v+dv
    if key == 'down':
        v = v-dv
        v = max(v,dv)
    if key == 'right':
        dv = 2*dv
    if key == 'left':
        dv = 0.5*dv
    return v,dv


###########################################################################
#Tests
###########################################################################
def testHandler():
    fig,ax = plt.subplots()
    x = np.linspace(0,1,100)
    ax.plot(x,x)
    ehd=eventHandler(fig,ax,path='/tmp/',name='test')
    plt.show()

if __name__=='__main__':
    #testHandler()
    print(green+__file__.split('/')[-1]+' success'+black)


from .displayStandards import*
#import matplotlib.pyplot as plt
from scipy import signal
from math import*

__all__ = ['Fourier','get_FFT','displayFourier','plot_fft2d',
'squareBis','unit_impulse','fftParams']
#### def : Main user interface
def Fourier(yfunc,Tf=100,dt=0.01,fopt='s',dOpt='ftc',pOpt='RIYP',
            xlims={'t':[],'f':[]},labOpts=['t','f'],
            Yth=lambda x:np.zeros((x.size)),figsize='12'):
    ''' where :
- fopt(Fourier opt)   : 's(sym) p(periodic)r(real)
- dOpt(display opt)   : 'f(freq)t(time)c(compare)d(fourierParams)s(spit Phase)
- pOpt(plot opt)      : 'R(real)I(imag)Y(abs)P(phase)
- Yth : lambda function to compare against
    '''
    symOpt,fft_func = 's' in fopt, [get_FFT,get_rFFT]['r' in fopt]

    #Signal and Fourier
    t = np.arange(0,Tf,dt) - Tf/2*symOpt
    y = yfunc(t)
    f,Y = fft_func(t,y,dt,'p' in fopt)

    #display
    df,fmax = fftParams(dt,Tf,dopt='d' in pOpt)
    if 't' in dOpt or 'f' in dOpt:
        nrows=2 if 'f' in dOpt and 't' in dOpt else 1
        ncols=2 if 's' in dOpt else 1
        fs=12;fonts=[fs]*5 #[fs,fsLeg,fsL,fsT ]
        fig,ax= create_fig(figsize=figsize,pad=0.75,rc=[nrows,ncols])
        axT,axF=ax
    if 't' in dOpt:
        displayFourier(t,y,labOpt=labOpts[0],opt=pOpt,ax=axT,xlims=xlims['t'],showOpt=0,fonts=fonts)
    if 'f' in dOpt:
        displayFourier(f,Y,labOpt=labOpts[1],opt=pOpt,ax=axF,xlims=xlims['f'],showOpt=0,fonts=fonts)
    if 'c' in dOpt:
        fig_c=compareFourier(f,Y,Yth(f),xlims=xlims['f'],showOpt=0)
        fig_c.show()
    if 't' in dOpt or 'f' in dOpt:
        fig.show()
    return t,y,f,Y

def get_FFT(t,y,dt=0,pOpt=False):
    ''' pOpt : periodic option => divide by sample size'''
    N,dt = t.size,t[1]-t[0] #shape[0]
    Y = np.fft.fftshift(np.fft.fft(y)*[dt,1./N][pOpt])
    f = np.fft.fftshift(np.fft.fftfreq(N,dt))
    #Y = np.fft.fft(y)*[dt,1./N][pOpt]
    #f = np.fft.fftfreq(N,dt)
    #f[:int(N/2)],Y[:int(N/2)]
    return f,Y

def get_iFFT(f,F,df=0,pOpt=False):
    ''' pOpt : periodic option hence divide by sample size'''
    N,df = f.size,f[1]-f[0] #shape[0]
    y = np.fft.fftshift(np.fft.ifft(F)*[df*N,1][pOpt])
    t = np.fft.fftshift(np.fft.fftfreq(N,df))
    return t,y

def get_rFFT(t,y,dt,Tf=0):
    N = t.size
    Y = 2*np.fft.rfft(y,N)/N
    f = np.fft.rfftfreq(N,dt)
    return f,Y

def fftParams(dt,Tf,dopt=1):
    ''' Tf : total length in time of the signal '''
    df,fmax   = 1./Tf, 0.5/dt
    if dopt==1:
        printParams(['Tf','dt', 'df','fmax'],[Tf,dt,df,fmax],s=4)
    if dopt==2:
        printParams(['sx','dx', 'dX','Xmax'],[Tf,dt,df,fmax],s=4)
    return df,fmax

###########################################################################
#### def : display
def plot_fft2d(f,F,xy=None,kxy=None,figsize='12',pad=None):
    '''
    f : function 2D NxN ndarray
    F : fft2D(f)
    xy,kxy : extent in real and reciprocal spaces
    '''
    #fig,((axm,axM),(axp,axP))=plt.subplots(nrows=2,ncols=2)
    #plt.tight_layout()
    fig,((axm,axM),(axp,axP)) = create_fig(figsize,pad,rc='22')
    plot_ax2D(axm,np.real(f),'$|f|$',xy)
    plot_ax2D(axp,angle(F),'$\phi(F)$',kxy,'P')
    plot_ax2D(axM,np.abs(F)**2,'$|F|^2$',kxy,'F')
    plot_ax2D(axP,np.abs(F),'$|F|$',kxy,'F')
    fig.show()
    return fig
def plot_ax2D(ax,f,title='',extent=None,opt='P',cmap='jet',org='lower'):
    A=f #/f.max()#;print(A)
    vmin,vmax=A.min(),A.max()
    if opt == 'P' : vmin,vmax = 0,2*pi
    if opt == 'F' : vmin=min(A.min(),0)
    ax.imshow(A,cmap=cmap,origin=org,extent=extent)#,vmax=vmax, vmin=vmin)
    ax.set_title(title)

def displayFourier(f,Y,labs=[],labOpt='f',opt='RIYPS',ax=None,xlims=[],showOpt=1,fonts=[]):
    plts = {'R':[f,np.real(Y),'c','$Re$'], 'Y':[f,np.abs(Y)   ,'b','$||$'],
            'I':[f,np.imag(Y),'m','$Im$'], 'P':[f,np.angle(Y) ,'r','$\Phi$'],
            'S':[f,np.power(np.abs(Y),2),'g','$S$']}
    labs = {'t':['$t$','$y$'],'f':['$f$','$F_y$'],
            'x':['$x$','$y$'],'k':['$k$','$F_y$']}[labOpt]
    if isinstance(ax,np.ndarray):
        plots = [plts[p] for p in opt if p in 'RI']
        stdDispPlt(plots,labs,ax=ax[0],xlims=xlims,showOpt=0,fonts=fonts)
        plots = [plts[p] for p in opt if p in 'YPS']
        stdDispPlt(plots,labs,ax=ax[1],xlims=xlims,showOpt=showOpt,fonts=fonts)
    else :
        plots = [plts[p] for p in opt if p in 'RIYPS']
        stdDispPlt(plots,labs,ax=ax,xlims=xlims,showOpt=showOpt,fonts=fonts)

def compareFourier(f,Y,Yth,xlims=[0,5,0,1],showOpt=1):
    plots = [[f,np.abs(Y),'g'   ,'$|Y|$'     ],
             [f,Yth      ,'b--' ,'$|Y|_{th}$']]
    fig_c,ax=stdDispPlt(plots,['$f$','$Y$'],xlims=xlims,showOpt=showOpt)
    return fig_c


###########################################################################
#### def : Misc
def angle(x,ee=1e-10) :
    '''Prevents round off errors from polluting the phase'''
    return ((np.abs(x)>ee)*np.angle(x)+
            ((abs(np.real(x))<ee) & (abs(np.imag(x)>ee)))*pi/2*np.sign(np.imag(x))+
            ((abs(np.imag(x))<ee) & (abs(np.real(x)>ee)))*pi*np.sign(np.real(x))+
            2*pi) %(2*pi)
def getFourierPair(t,opt=0, f0=1.0,tau0=0.1,t0=0.0,phi0=0.0):
    fy = {'cos':lambda t:np.cos(2*pi*f0*(t-t0)+phi0),
          'sin':lambda t:np.sin(2*pi*f0*(t-t0)+phi0),
          'exp':lambda t:np.cos(2*pi*f0*(t-t0)+phi0)*np.exp(-np.power((t-t0)/(tau0/2),2)),
          'gau':lambda t:np.exp(-tau0*np.power(t,2))}
    y = fy[opt](t)
    return y
def get_extents(a,npts):
    dx=a/npts
    xy = [0,a]*2
    kxy = np.fft.fftshift(np.fft.fftfreq(npts,dx))[[0,-1]].tolist()*2
    return xy,kxy
def unit_impulse(n,i):
    y = np.zeros((n));y[i] = 1
    return y
def square(t,T,nT):
    N,nT0 = t.size,nT/2
    y = np.zeros((int(N/2)))
    t0=t+t[-1];t0=t0[t0<T*(nT0+0.5*nT%2)]
    if nT%2:
        y[t0%T>0.75*T] = 1
        y[t0%T<0.25*T] = 1
    else:
        y[t0%T>0.5*T] = 1
    if not N%2:
        y = np.concatenate((np.flipud(y),y))
    else:
        y = np.concatenate((np.flipud(y),y,np.array([0])))
    return y
def squareBis(t,T,nT,duty=0.5,applyScale=True):
    scale = [1,0.5][applyScale]
    if nT==np.Inf:
        y = scale*signal.square(2*pi*(t+0.25*T)/T,duty)+scale*applyScale
    else:
        if nT%2:
            y = scale*signal.square(2*pi*(t+0.5*duty*T)/T,duty)+scale*applyScale
        else:
            y = scale*signal.square(2*pi*(t+(duty+0.5*(1-duty))*T)/T,duty)+scale*applyScale
        y[t<-T*nT*0.5]=0;y[t>+T*nT*0.5]=0;
    return y

###########################################################################
#### def : test
###########################################################################
def testDisplayFourier():
    t = np.linspace(0,10,100)
    displayFourier(t,np.exp(2j*pi*t),opt='RIYP')
    displayFourier(t,np.exp(2j*pi*t),opt='RI')
    displayFourier(t,np.exp(2j*pi*t),opt='YP')
    displayFourier(t,np.exp(2j*pi*t),opt='Y')

def testFourier(y='gauss',params={'f0':0.0,'tau0':1.0,'t0':0,'phi0':0}):

    f0,tau0,t0,phi0 = [params.get(key) for key in ['f0','tau0','t0','phi0']]
    if y=='basic':
        yfunc,Yfunc = lambda t:np.exp(-t*t), lambda f:sqrt(pi)*np.exp(-pi**2*f*f)
        Fourier(yfunc,Tf=100,dt=0.01,fopt='s',Yfunc=Yfunc,dOpt='ft',pOpt='Y',
                xlims={'t':[-4,4,0,1],'f':[0,1,0,2]})
    if y=='gauss':
        Df  = 1.0/(pi*tau0/2);
        yfunc = lambda t:np.exp(2j*pi*f0*(t-t0)+1j*phi0)*np.exp(-np.power((t-t0)/(tau0/2),2))
        Yfunc = lambda f:sqrt(pi)*tau0/2*np.exp(-np.power((f-f0)/(Df),2))
        Fourier(yfunc,Tf=100,dt=0.01,fopt='s',Yfunc=Yfunc,dOpt='ftc',pOpt='RIY',
                xlims={'t':[-4,4,-1,1],'f':[0,4,0,1.5]})
    elif y=='cos':
        yfunc = lambda t:np.cos(2*pi*f0*(t-t0)+phi0)
        Yfunc = lambda f:unit_impulse(f.size,np.argmin(np.abs(f-f0)))
        Fourier(yfunc,Tf=100,dt=0.01,fopt='rp',Yfunc=Yfunc,dOpt='ftc',pOpt='R',
                xlims={'t':[0,5,-1,1],'f':[0,4,0,1]})
    elif y=='expi':
        yfunc = lambda t:np.exp(2j*pi*f0*(t-t0)+1j*phi0)
        Fourier(yfunc,Tf=100,dt=0.01,fopt='p',dOpt='ft',pOpt='RIY',
                xlims={'t':[0,5,-1.1,1.1],'f':[f0-0.1,f0+0.1,-0.1,1.1]})
def test_square():
    t,T = np.linspace(-10,10,1000),2
    Ns = [2,5,np.Inf]
    cs = getCs('jet',len(Ns))
    plots = [[t,squareBis(t,T,Ns[i],0.25,applyScale=False),cs[i],'$%s$' %Ns[i]] for i in range(len(Ns))]
    stdDispPlt(plots,['$t$','$y$'])

if __name__ =='__main__':
    #testDisplayFourier()
    #testFourier('basic')
    #testFourier('gauss',{'f0':0.0,'tau0':1.5,'t0':0.0,'phi0':0.0})
    #testFourier('cos'  ,{'f0':0.0,'tau0':1.5,'t0':0.0,'phi0':pi/3})
    #testFourier('expi' ,{'f0':2.0,'tau0':1.5,'t0':0.0,'phi0':pi/3})
    #test_square()
    print('FourierTest compiled');print("")



import numpy as np
import matplotlib.pyplot as plt
from .displayStandards import*
from .glob_colors import*

def y_str(v):
    if isinstance(v,complex) or isinstance(v,np.complex128):
        return '%.3f+j%.3f' %(v.real,v.imag)
    elif isinstance(v,np.ndarray):
         return '['+' '.join(map(lambda s:'%.2E' %s,v))+']'
    else:
        return '%.3f' %v
def newton_recursive(f,df,x,y0,cond=None,Ncut=5,npts=3,tol=1e-5,nMax=100,v=1):
    ''' find roots of f(y,x) with Newton for fixed values of x (works both for 1D and N-D functions F):\n
    - x : ndarray - set of values at which netwton is computed
    - y0 : initial guess
    - cond : cond(x0,x1,y0,y1) functor - condition for accepting the solution (default accepted if f0<tol)
    - Ncut,npts : max cutback level and number of intermediate pts per cutback
    - v : verbose level(1 iterations for x, 2 cutback levels)
    - tol,nMax : Newton arguments
    '''
    if isinstance(y0,complex) or isinstance(y0,np.complex128):
        is_nd = 1
        y = np.zeros(x.shape,dtype=complex)
    elif isinstance(y0,list) or isinstance(y0,np.ndarray) :
        y0 = np.array(y0)
        is_nd = 2
        y = np.zeros((x.shape[0],y0.size),dtype=float)
    else:
        is_nd = 0 # default newton1D
        y = np.zeros(x.shape,dtype=float)
    newton_p = [newtonRaphson1D,newton_cplx,newton][is_nd]
    if not cond :
        cond = lambda x0,x1,y0,y1:True
        if v>1 : print('Newton recursive defaut condition')
    #check initial guess
    y0,f0  = newton_p(f,df,y0,(x[0],),tol,nMax=10000,v=1)
    if f0<tol :
        y[0] = y0
        if v :
            print(yellow+'\tNewton Raphson %s recursion log' %(['1D','cplx','ND'][is_nd]) + black)
            print(yellow+'  i : x     y     \n'+'-'*20+black)
            print(yellow+'%3d : %s %s' %(0,y_str(x[0]),y_str(y0)) +black)
    else :
        print(red+'bad initial guess :')
        print('x0=%s,y0=%s,f0=%E' %(y_str(x[0]),y_str(y0),f0))
        print('exiting'+black);return x
    #iterate
    i,valid = 1,f0<tol
    while i<x.size and valid:
        y0,valid = _newton_rec(0,f,df,x[i-1:i+1],y0,cond,newton_p,Ncut,npts,tol,nMax,v>1)
        if v : print(yellow+'%3d : %s %s' %(i,y_str(x[i]),y_str(y0)) +black)
        y[i],i = y0,i+1
    if not valid and v :
        print(red+'convergence issue : last point (invalid)'+black)
        print('x0=%s,y0=%s,f0=%E' %(y_str(x[i]),y_str(y[i]),f0))
    return y

def _newton_rec(n,f,df,x,y0,cond,newton_p,Ncut,npts,tol,nMax,v):
    ''' Solve recursively with Newton until found
    '''
    y1,valid = y0,False
    if n<Ncut:
        i,valid = 1,True
        if n and v: print(red+'cutback level %d for x=%.2f : ' %(n,x[-1]) +black,'new range : ',x)
        while i<x.size and valid :
            x0,x1,y0 = x[i-1],x[i],y1
            if n and v : print(red+'%3d : %.6f' %(i,x1)+black)
            y1,f1  = newton_p(f,df,y0,(x1,),tol,nMax,v=1)
            valid = f1<tol and cond(x0,x1,y0,y1)
            if not valid:
                y1,valid = _newton_rec(n+1,f,df,np.linspace(x0,x1,npts),y0,cond,newton_p,Ncut,npts,tol,nMax,v)
                # valid &= cond(x0,x1,y0,y1)
            i += 1
        if n and v :print([red,green][valid]+'cutback level %d %s' %(n,['fail','success'][valid])+black)
    return y1,valid

########################################################################
# Newton algos
def newton(f,df,x0,args=(),tol=1e-5,nMax=100,v=0):
    if isinstance(df,int):
        return newtonfdf(f,x0,args,tol,nMax,v)
    else:
        return newtonf(f,df,x0,args,tol,nMax,v)


def newtonfdf(fdf,x0,args=(),tol=1e-5,nMax=100,v=0):
    x,i = x0,0
    fx,dfx = fdf(x,*args)
    fxN = np.linalg.norm(fx)
    while fxN>=tol and i<nMax:
        fx,dfx = fdf(x,*args)
        # print(x,fxN,fx,dfx)
        dx = np.linalg.solve(dfx,fx)
        x -= dx
        fxN = np.linalg.norm(fx)
        i = i+1
    return [x,fxN,i][:v+1]

def newtonf(f,df,x0,args=(),tol=1e-5,nMax=100,v=0):
    x,i = x0,0
    fx,dfx = f(x,*args)
    fxN = np.linalg.norm(fx)
    while fxN>=tol and i<nMax:
        fx,dfx = f(x,*args),df(x,*args)
        # print(x,fxN,fx,dfx)
        dx = np.linalg.solve(dfx,fx)
        x -= dx
        fxN = np.linalg.norm(fx)
        i = i+1
    return [x,fxN,i][:v+1]

def newton_cplx(f,df,z0,args=(),tol=1e-7,nMax=100,v=0):
    ''' Newton for a complex valued function f(z)
    - f,df : functor f(z), df(z) - complex valued function and its derivative
    - z0 : complex - initial guess
    '''
    F  = lambda z:np.abs(f(z[0]+1J*z[1],*args))**2
    dF = lambda z:grad_fz2(f(z[0]+1J*z[1],*args), df(z[0]+1J*z[1],*args))

    z,i     = [z0.real,z0.imag],0
    fz      = F(z)
    while fz>=tol and i<nMax :
        dfz = dF(z)
        # print(z[0]+1J*z[1],fz,dfz)
        z -= dfz*fz/np.linalg.norm(dfz)**2
        fz = F(z)
        i = i+1
    out = [z[0]+1J*z[1],fz,i]
    return out[:v+1]
def grad_fz2(fz,dfz):
    u,v = fz.real,fz.imag
    ux,vx = dfz.real,dfz.imag
    return np.stack([2*u*ux + 2*v*vx,-2*u*vx+2*v*ux])


def newtonRaphson1D(f,df,x0,args=(),tol=0.01,nMax=10,v=0):
    x,i     = x0,0
    #argsx   = (x,)+args
    fx      = f(x,*args)
    while abs(fx)>=tol and i<nMax :
        print(x,fx,df(x,*args))
        x = x - fx/df(x,*args)
        #argsx = (x,)+args
        fx = f(x,*args)
        i = i+1
    X = [x,abs(fx),i]
    return X[:v+1]

############################################################
#### def : Efficient
def newtonRaphsonEff(fdf,x0,tol=0.01,nMax=10):
    x,i = x0,0
    fx,dfx = fdf(x)
    fxN = np.linalg.norm(fx)
    while fxN>=tol and i<nMax:
        fx,dfx = fdf(x)
        dx = np.linalg.solve(dfx,fx)
        x = np.subtract(x,dx)
        fxN = np.linalg.norm(fx)
        i = i+1
    return x

def newtonRaphson1DEff(fdf,x0,tol=0.01,nMax=10):
    x,i  = x0,0
    fx,dfx = fdf(x)
    while abs(fx)>=tol and i<nMax:
        x = x - fx/dfx
        i = i+1
        fx,dfx = fdf(x)
    return x

############################################################
#### def : Debug
def newtonRaphson_dbg(f,df,x0,args=(),tol=0.01,nMax=10,v=0):
    x,i,xi = x0,0,[]
    argsx = (x,)+args
    fx = np.linalg.norm(f(*argsx))
    while fx>=tol and i<nMax:
        xi += [np.array(x)]; print(xi)
        dfx = df(*argsx)
        dfinv = np.linalg.inv(dfx)
        dx = dfinv.dot(f(*argsx))#np.linalg.solve(df(*argsx),f(*argsx))
        print(red,x,black);print(dfx);print(dfinv);print(f(*argsx));print(dx);print(fx)
        x -= dx # np.subtract(x,dx)
        argsx = (x,)+args
        fx = np.linalg.norm(f(*argsx))
        i = i+1
    xi+=[np.array(x)]
    #X = [x,fx,i]
    return np.array(xi) #X[:v+1]
def newtonRaphsonDebug(f,df,x0,tol=0.01,nMax=10):
    x = x0
    fx = np.linalg.norm(f(x))
    print(fx);plt.plot(x,'o-');
    i = 0
    while fx>=tol and i<nMax:
        dx = np.linalg.solve(df(x),f(x))
        x = np.subtract(x,dx)
        fx = np.linalg.norm(f(x))
        i = i+1
        print(fx);plt.plot(x,'o-',color=(1-fx/tol,0,0));
    return x,fx,i
def newtonRaphson1Ddebug(f,df,x0,X,tol=0.01,nMax=10,logOpt=0):
    fX      = np.log10(abs(f(X))) if logOpt else f(X)
    plots   = [[X ,fX ,'b-','$f(x)  $']]

    dispIntermediateSol(plots,x0,f,'r',tol,logOpt)
    x,i = x0,0
    while abs(f(x))>=tol and i<nMax:
        x = x - f(x)/df(x)
        i = i+1
        dispIntermediateSol(plots,x,f,'b',tol,logOpt)
    dispIntermediateSol(plots,x,f,'g',tol,logOpt)

    stdDispPlt(plots,['x','fx'],xlims=['y',-100.,100.])
    return x,f(x),i


############################################################
#### def : Misc
def dispIntermediateSol(plots,x,f,c,tol,logOpt=0):
    fx = f(x)
    if logOpt:
        logTol  = abs(np.log10(tol))
        logFx   = logTol if fx<=tol else abs(np.log10(abs(fx)))
        #c       = abs(pow(logFx/logTol,0.3))
        fx      = abs(logFx)
    print(x, fx)
    plots.append([x,fx,c+'o',''])
    return plots

def f_cplx(f,z,z0):
    z = z[0]+1J*z[1]
    fz = f(z,z0)
    return np.array([fz.real,fz.imag])

def j_cplx(df,z,z0):
    z = z[0]+1J*z[1]
    df_z,dfz_ = df(z,z0),df(np.conj(z),z0)
    jac = [ [df_z+dfz_      ,  1J*(df_z+dfz_)],
            [-1J*(df_z-dfz_),       df_z-dfz_], ]
    jac = 0.5*np.array(jac)
    print('real :\n' ,jac.real)
    print('imag :\n' ,jac.imag)
    return jac #.real

############################################################
##### def : Tests
def _newtonExample(debugOpt=0):
    print('\t')                         ;print('Newton Raphson ' )
    f  = lambda x:[pow(x[0],2)-1,x[1]]  ;print("f(x)  = [x(0)^2-1,x(1)] ")
    df = lambda x:[[2*x[0]  ,0],
                   [0       ,1]]        ;print("df(x) = [[2*x(0)^2,0] \n"
                                           "         [0       ,1]]")
    x0 = [2,2]                          ;print("x0 = [2,2]")
    x = newtonRaphson(f,df, x0)         ;print("x  = " + valsToStr(x[0]))
    if debugOpt:
        X = np.meshgrid(np.linspace(0,2,10),np.linspace(0,2,10))
        fX = f0(X)
        plt.plot();plt.show()
def _newtonExample1D(debugOpt=0,logOpt=0):
    x0 = [-2.,2.5]
    print('\t')                     ;print('Newton Raphson 1D ' )
    f  = lambda x:pow(x,2)-1        ;print('f  = x^2-1')
    df = lambda x:2*x               ;print('df = 2*x')
    x  = newtonRaphson1D(f,df,x0[0]);print('x0=%-3.2f => x =%.2f'  %(x0[0],x[0]))
    x  = newtonRaphson1D(f,df,x0[1]);print('x0=%-3.2f => x =%.2f'  %(x0[1],x[0]))

    if debugOpt:
        X = np.linspace(-3,3,1002)
        newtonRaphson1Ddebug(f,df,x0[1],X,tol=pow(10.,-12),nMax=10,logOpt=logOpt)

def _newtonExample1DdebugQuick():
    f  = lambda x:pow(x,3)-1
    df = lambda x:3*pow(x,2)
    X  = np.linspace(-3,3,1000)
    x  = newtonRaphson1Ddebug(f,df,-10.,X,tol=pow(10.,-5),nMax=20,logOpt=0);
    print(x)

######################################################################
if __name__=='__main__':
    import numpy as np
    #newtonExample(0)
    #newtonExample1D(debugOpt=1,logOpt=0)
    #newtonExample1DdebugQuick()
    print('succesful compilation')



import matplotlib.pyplot as plt

def bisect(f,a,b,args=(),rtol=1e-8, Nmax=100,v=1) : #,full_output=0):
    ''' Bisection method
    v : 1:warning if tol not met, 2:full_output with f0
    '''
    x0,N = float(a+b)/2,0
    args0 = (x0,)+args;
    f0 = abs(f(*args0))
    while f0>rtol and N<Nmax:
        x1 = (0.75*a + 0.25*b)
        x2 = (0.25*a + 0.75*b)
        args1,args2 = (x1,)+args, (x2,)+args
        f1,f2 = abs(f(*args1)),abs(f(*args2))
        if f1>f2 :
            a = x1
        else:
            if f2>f1:
                b = x2
            else:
                a = x1
                b = x2
        x0,N = (a+b)/2, N+1
        args0 = (x0,)+args
        f0 = abs(f(*args0))
    if v==1 :
        if f0>rtol:print('Warning tolerance not achieved f(x0)=%E,Nmax=%d'  %(f0,Nmax))
    if v==2 : return x0,f0
    return x0


def bissectionPos(f,a,b,tol=0.001, Nmax=100):
    x0,N = (a+b)/2,0
    while abs(f(x0))>tol and N<Nmax:
        #print abs(f(x0)),x0
        if  f(x0)>0:
            b = x0
        else:
            a = x0
        x0 = float(a+b)/2
        N+=1
    return x0,f(x0)

def bissection(f,a,b,tol=0.001, Nmax=100, plotOpt=0):
    x0 = float(a+b)/2
    N  = 0
    while f(x0)>tol and N<Nmax:
        x1 = (0.75*a + 0.25*b)
        x2 = (0.25*a + 0.75*b)
        if      f(x1)>f(x2):
            a = x1
        else:
            if f(x2)>f(x1):
                b = x2
            else:
                a = x1
                b = x2
        x0 = (a+b)/2
        N+=1
    return x0,f(x0)

def bissection2(f,a,b,tol=0.001, Nmax=100, plotOpt=0):
    x0 = float(a+b)/2
    N  = 0
    while abs(a-b)>tol and N<Nmax:
        x1 = (0.75*a + 0.25*b)
        x2 = (0.25*a + 0.75*b)
        if      f(x1)>f(x2):
            a = x1
        else:
            if f(x2)>f(x1):
                b = x2
            else:
                a = x1
                b = x2
        x0 = (a+b)/2
    return x0


def bissectionDebug(f,a,b,tol=0.001, Nmax=100, plotOpt=0):
    x0 = (a+b)/2
    N  = 0
    f0 = f(x0)
    f0s = [f0]
    while f0>tol and N<Nmax:
        x1 = (0.75*a + 0.25*b)
        x2 = (0.25*a + 0.75*b)
        if  f(x1)>f(x2):
            a = x1
        elif f(x2)>f(x1):
            b = x2
        else:
            a = x1
            b = x2
        x0 = (a+b)/2


        N  = N+1
        f0 = f(x0)
        f0s.insert(N,f0)
        #print x0

    if N==Nmax:
        print('''warning :'
                  bissection did not get to the required tolerance tol=%f
                  within the limited number of iterations Nmax=%d
              ''' %(tol,Nmax))
    if plotOpt:
        plt.plot(range(len(f0s)),f0s)
        plt.show()
    return x0,f(x0)

if __name__=='__main__':
##    f = lambda x:pow(x-1.0,2)
##    x0,f0 = bissection(f,0,1,0.001,12)
##    print('bissection result : %f with residual %f ' %(x0, f0))
##    x0,f0 = bissectionDebug(f,0,1,0.00001,100,1)

    print('Bissection compiled')



import setuptools

#with open("latex/README.md", "r") as fh:
#    long_description = fh.read()

setuptools.setup(
    name="TarikDrevonUtils",
    version="1.0.8",
    author="Tarik Ronan Drevon",
    author_email="ronandrevon@gmail.com",
    description="some display utilities and others",
    long_description='', #long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_data={'utils/materials':['materials*.pkl']},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License ",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'matplotlib','numpy','scipy','colorama','pandas','pillow','PyQt5'],
)



import numpy as np
import warnings
#import __builtin__
#defaultPath = 'C:\Users\ronan\Documents\Silvaco\PhysicsCode\Common'

def print_args(fmt0='%.3f',**kwargs):
    '''example : print_params('%.2E',a=1e18,b=1e19)'''
    print_params(kwargs,fmt0)

def print_params(params_d,fmt0='%.3f'):
    ''' print params in a function parsing manner
    - fmt0 will be used for print formatting (so forbidden entry)
    - params_d : dictionary of values to print
    '''
    msg = ''
    spacing = max([len(getParamUnit(k)[0]) for k in params_d.keys()])
    fmt = '%%-%ds = %s %%s\n' %(spacing,fmt0)
    for k,v in params_d.items():
        p,unit = getParamUnit(k)
        msg+=fmt %(p,v,unit)
    print(msg[:-1])

#### def : Main user interface
# load and save arrays
def saveParams(name,dat, datType):
    path = checkPath()
    filepath = '%s/dat/%s_%s' %(path,name,datType)
    np.save(filepath,dat)
    print('%s_%s successfully saved in %s' %(name,datType, path))

def loadParams(name,datType):
    print(' Loading %s_%s ' %(name,datType))
    path = checkPath()
    filepath = '%s/dat/%s_%s.npy' %(path,name,datType)
    dat = np.load(filepath)
    return dat

def printParams(paramNames,params,f='f',prec=3,s=5):
    d  = {'f':0,'E':1}[f]
    xp = lambda x:[
        ['%.0f' %x,'%.1f' %x,'%.2f' %x,'%.3f' %x,'%.4f' %x ,'%.5f' %x,
         '%.6f' %x,'%.7f' %x,'%.8f' %x,'%.9f' %x,'%.10f' %x],
        ['%.0E' %x,'%.1E' %x,'%.2E' %x,'%.3E' %x,'%.4E' %x ,'%.5E' %x,
         '%.6E' %x,'%.7E' %x,'%.8E' %x,'%.9E' %x,'%.10E' %x]][d][prec]
    xs = lambda x:[
        '%-0s' %x,'%-1s' %x,'%-2s'%x,'%-3s'%x,'%-4s'%x,'%-5s'%x,'%-6s'%x,'%-7s'%x,'%-8s'%x,'%-9s'%x,
        '%-10s'%x,'%-11s'%x,'%-12s'%x,'%-13s'%x,'%-14s'%x,'%-15s'%x,'%-16s'%x,'%-17s'%x,'%-18s'%x,'%-19s'%x,
        '%-20s'%x][s]

    for paramUnit,p in zip(paramNames,params):
        paramName,unit = getParamUnit(paramUnit)
        pName = xs(paramName)
        if isinstance(p,int):
            print('%s = %d %s'          %(pName,p, unit))
        elif isinstance(p,float):
            print('%s = %s %s'          %(pName,xp(p),unit))
        elif isinstance(p,complex):
            print('%s = %s + j%s %s'    %(pName,xp(p.real),xp(p.imag),unit))

def getParamUnit(paramUnit):
    paramSplit = paramUnit.split('(')
    param = paramSplit[0]
    unit  = paramSplit[1].split(')')[0] if len(paramSplit)>1 else ''
    return param,unit

def printMatrices(Mat,MatNames,prec=2):
    np.set_printoptions(precision=prec)
    np.set_printoptions(suppress=False)
    for i in range(len(Mat)):
        print('%s : ' %MatNames[i])
        print(Mat[i])

#### handle complex for optimization
cplx2list = lambda c:[c.real,c.imag]
list2cplx = lambda c:c[0]+1j*c[1]

#### Misc :
# def checkPath():
    # if not hasattr(__builtin__, "path"):
        # warnings.warn( "Variable 'path' not defined. default path used instead :"
                       # "\n %s" %defaultPath)
        # path = defaultPath.replace('\\' ,'/').replace('\r' ,'/r').replace('\t' ,'/t')
    # else:
        # path = __builtin__.path
    # return path

def initVector(M,N):
    V = [np.zeros((N)) for i in range(M)]
    return V

def valsToStr(v,n=2,f='E'):
    if f == 'E':
        if n==1:
            vals = ['%.1E' %(val) for val in v];
        elif n==2:
            vals = ['%.2E' %(val) for val in v];
        else:
            vals = ['%.3E' %(val) for val in v]
    elif f=='f' :
        if n==1:
            vals = ['%.1f' %(val) for val in v];
        elif n==2:
            vals = ['%.2f' %(val) for val in v];
        else:
            vals = ['%.3f' %(val) for val in v]
    elif f=='d' :
        vals = ['%d' %(val) for val in v]
    else :
        return ''
    return  '[' + ', '.join(vals) + ']'

def printVals(v,f='E',prec=3):
    values = ''
    d  = {'f':0,'E':1}[f]
    xp = lambda x:[
        ['%.0f' %x,'%.1f' %x,'%.2f' %x,'%.3f' %x,'%.4f' %x ,'%.5f' %x,
         '%.6f' %x,'%.7f' %x,'%.8f' %x,'%.9f' %x,'%.10f' %x],
        ['%.0E' %x,'%.1E' %x,'%.2E' %x,'%.3E' %x,'%.4E' %x ,'%.5E' %x,
         '%.6E' %x,'%.7E' %x,'%.8E' %x,'%.9E' %x,'%.10E' %x]][d][prec]
    for val in v:
        if isinstance(val,int):
            values += '%-12d' %(val)
        elif isinstance(val,str):
            values += '%-12s' %(val)
        else:
            values += '%-12s' %(xp(val))
    print(values)

def printKeys(k):
    keys, line = ['','']
    strLine = '-'*12
    for key in k:
        keys   = keys + '%-12s' %(key)
        line   = line + strLine
    print(keys)
    print(line)



#################################################################################
#### def : test
#################################################################################
def testArraySaveLoadPrint():
    name ='test'
    dat = np.random.rand(5)
    saveParams(name,dat,'array')
    datLoaded = loadParams(name,'array')
    params = ['a%d' %i for i in range(len(dat))]
    printParams(params,dat)

def testPrintParams():
    print('\t prec=3,s=2,f="E"')
    printParams(['i','f','c'],[1,2.0,1+1j],prec=3,s=2,f='E')
    print('\t prec=2,s=8')
    printParams(['integer','float','complex'],[1,2.0,1+1j],prec=2,s=8)
    print('\t prec=2,s=8')
    printParams(['test','gain(cm-1)','E(meV)','T(K)'],[1,100,1,0.0,300])

def testgetParamUnit():
    paramUnits = ['param','param(unit)','param(unit0']
    for paramUnit in paramUnits:
        print(getParamUnit(paramUnit))

if __name__=='__main__':
    #testArraySaveLoadPrint()
    #testPrintParams()
    #testgetParamUnit()
    print('InputOutput defined')
    print("ok")



from utils import*
import utils.displayStandards as dsp
import importlib as imp
imp.reload(dsp)


def _testStdDispPlt():
    nPts = 10
    x1 = np.linspace(-2,2,nPts);y1 = pow(x1,1)
    x2 = np.linspace(-2,2,nPts);y2 = pow(x2,2)
    x3 = np.linspace(-2,2,nPts);y3 = pow(x3,3)
    plots = [[x1,y1,'r+-.','$x$'  ,1],
             [x2,y2,'gs--','$x^2$',2],
             [x3,y3,'bo--','$x^3$',3]]
    xlabs = ['$x$','$y$']
    stdDispPlt(plots,xlabs, axPos=[0.12, 0.12, 0.8, 0.8], c=['k','b'],
               fonts=[30,25,12,25], texts=[[0.05,0.2,'O','m']])
def _testStdDispPlt_b():
    x = np.linspace(0,1,10)
    plots       = [x,x,'b','x']
    plotsList   = [plots]
    stdDispPlt(plots, showOpt=0)
    stdDispPlt(plotsList)
def _testStdDispPlt_c():
    x = np.linspace(0,1,10)
    plots       = [x,x,[(1.0,0.5,0.2),'s-.'],'x']
    stdDispPlt(plots, showOpt=1)
def _testStdDispPlt_subplots():
    fig,ax = plt.subplots(nrows=1,ncols=2)
    x = np.linspace(0,1,10)
    stdDispPlt([x,2*x,'b','$y1$'],['$x$','$y$'], ax=ax[0],axPos=11, showOpt=0)
    stdDispPlt([x,4*x,'r','$y2$'],['$t$','$y$'], ax=ax[1],axPos=12, showOpt=1,fullsize=True)

def _testAddyAxis():
    nPts = 10
    x1   = np.linspace(-2,2,nPts);y1 = pow(x1,1)
    plotsAx1 = [x1,+y1,'r' ,'']
    plotsAx2 = [x1,-y1,'b','']
    fig,ax = stdDispPlt(plotsAx1,['$x$','$y$'],c=['k','r'],legOpt=0,showOpt=0)
    addyAxis(fig,ax,plotsAx2,'$-y$',c='b', legOpt=0, showOpt=1)
def _testAddxAxis():
    x1 = np.linspace(-2,2,100);y1 = pow(x1,2)
    x2 = np.linspace(-1,1,100);y2 = pow(x2,2)
    plotsAx1 = [x1,y1,'r' ,'$x^2$']
    plotsAx2 = [x2,y2,'b' ,'$x^2$']
    fig,ax = stdDispPlt(plotsAx1,['$x$','$y$'],c=['k','k'],legOpt=0,showOpt=0)
    addxAxis(ax,plotsAx2,'$x_2$',c='b', showOpt=1)

def _testChangeAxLim():
    fig,ax = plt.subplots();plt.plot(range(5),range(5))
    print(changeAxesLim(ax,0.05,xylims=[]))
    print(changeAxesLim(ax,0.05,xylims=[0,1.,0,2.]))
    print(changeAxesLim(ax,0.05,xylims=['x',0,1.]))
    print(changeAxesLim(ax,0.05,xylims=['y',0,2.]))
def _test_get_axPos():
    cmd ='get_axPos(1)' ; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos(11)'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos(21)'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos(0)' ; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos([])'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos([0.1, 0])'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos([0.1, 0.1, 0.8, 0.8])'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
    cmd ='get_axPos((0.1, 0.1, 0.8, 0.8))'; print('%-20s: axPos=%s' %(cmd,str(eval(cmd))))
def _testGetSymCurve():
    x = np.linspace(0,np.pi,100)
    x1,x2 = x,x
    y1,y2 = np.cos(x1),np.sin(x2)
    x1,y1 = getSymCurve(x1,y1,1)
    x2,y2 = getSymCurve(x2,y2,-1)
    stdDispPlt([[x1,y1,'b','$cos$'],[x2,y2,'r','$sin$']])
def _testGetCML(s):
    c,m,l = getCML(s)
    print('color, marker, linestyle')
    print(c,m,l)
def _test_scat() :
    x,y = np.random.rand(2,10)
    c   = ['b']*10
    s   = [5]*10
    ct  = [(0.2,0.3,0.1)]*10
    stddisp(scat=[x,y,s,c]          ,opt='q',legOpt=0)
    stddisp(scat=[x,y,s[0],c[0]]    ,opt='q',legOpt=0)
    stddisp(scat=[x,y,c]            ,opt='q',legOpt=0)
    stddisp(scat=[x,y,np.array(c)]  ,opt='q',legOpt=0)
    stddisp(scat=[x,y,ct]           ,opt='q',legOpt=0)
    stddisp(scat=[x,y,ct[0]]        ,opt='p',legOpt=0)
def _test_stddisp():
    nPts = 10
    x1 = np.linspace(-2,2,nPts);y1 = pow(x1,1)
    x2 = np.linspace(-2,2,nPts);y2 = pow(x2,2)
    x3 = np.linspace(-2,2,nPts);y3 = pow(x3,3)
    plots = [[x1,y1,'r+-.','$x$'  ,1],
             [x2,y2,'gs--','$x^2$',2],
             [x3,y3,'bo--','$x^3$',3]]
    pol=matplotlib.patches.Polygon([[0,0],[0,1],[1,1]],label='$pol$')
    coll = matplotlib.collections.PatchCollection([pol],'y',alpha=0.3,linewidth=2,edgecolor='g')
    stddisp(plots=plots,colls=[coll],scat=[x1,y1,10,'b'],
            texts=[[0.05,0.2,'This is text','m']],
            labs=['$x$','$y$'], c=['k','b'],pad=2,
            fonts={'text':30,'title':15,'lab':15,'tick':10,'leg':'20'},
            legElt=[pol],opt='p')



def test_display_solutions():
    disp_d = {
        'x1':('get_x1','b',r'$x_1$'),
        'x2':('get_x2','c',r'$x_2$'),
        'y1':('get_y1','r',r'$y_1$'),
        'y2':('get_y2','m',r'$y_2$'),
        }
    class C1:
        def get_x1(self):return np.linspace(0,1,10)
        def get_x2(self):return np.linspace(0,2,10)
        def get_y1(self):return 1*self.get_x1()
        def get_y2(self):return 2*self.get_x2()
    class C2:
        def get_x1(self):return np.linspace(0,1.5,20)
        def get_x2(self):return np.linspace(0,3.0,20)
        def get_y1(self):return 0.5*self.get_x1()
        def get_y2(self):return 3*self.get_x2()
    c1,c2 = C1(),C2()
    objs=[(c1,'ko','$c_1$'), (c2,'ks','$c_2$')]

    dsp.display_solutions(disp_d,objs,key_pair=('x1','y1'),help=0,lw=2)
    dsp.display_solutions(disp_d,objs,key_pair=('x2',['y1','y2']),help=0,lw=2)

if __name__=='__main__':
    plt.close('all')
    #_testGetCML('b:.<')
    #_testStdDispPlt()
    #_testStdDispPlt_b()
    #_testStdDispPlt_c()
    #_testStdDispPlt_subplots()
    #_testAddxAxis()
    #_testAddyAxis()
    #_testChangeAxLim()
    #_testGetSymCurve()
    #_test_get_axPos()
    #_test_scat()
    #_test_stddisp()
    #print(getCML(np.random.rand(3))
    test_display_solutions()
